[![](https://www.r-pkg.org/badges/version/scoper)](https://cran.r-project.org/package=scoper)
[![](http://cranlogs.r-pkg.org/badges/grand-total/scoper)](https://www.r-pkg.org/pkg/scoper)
[![](https://cranlogs.r-pkg.org/badges/scoper)](https://www.r-pkg.org/pkg/scoper)
[![](https://img.shields.io/static/v1?label=AIRR-C%20sw-tools%20v1&message=compliant&color=008AFF&labelColor=000000&style=plastic)](https://docs.airr-community.org/en/stable/swtools/airr_swtools_standard.html)
[![](https://img.shields.io/badge/github-repo-blue?logo=github&link=https://github.com/immcantation/scoper)](https://github.com/immcantation/scoper)

SCOPer
-------------------------------------------------------------------------------

SCOPer (Spectral Clustering for clOne Partitioning) provides a computational framework for the identification of B cell 
clonal relationships from Adaptive Immune Receptor Repertoire sequencing 
(AIRR-Seq) data. It includes methods for assigning clonal identifiers using
sequence identity, hierarchical clustering, and spectral clustering.
SCOPer is part of the [Immcantation](http://immcantation.readthedocs.io) 
analysis framework.

Contact
-------------------------------------------------------------------------------

If you need help or have any questions, please contact the [Immcantation Group](mailto:immcantation@googlegroups.com).

If you have discovered a bug or have a feature request, you can open an issue using the [issue tracker](https://github.com/immcantation/scoper/issues).

To receive alerts about Immcantation releases, news, events, and tutorials, join the [Immcantation News](https://groups.google.com/g/immcantation-news) Google Group. [Membership settings](https://groups.google.com/g/immcantation-news/membership) can be adjusted to change the frequency of email updates.
