# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

pairwiseMutMatrixRcpp <- function(informative_pos, mutMtx, motifMtx) {
    .Call(`_scoper_pairwiseMutMatrixRcpp`, informative_pos, mutMtx, motifMtx)
}

countSeqsWithInvalidBases_rcpp <- function(seqs) {
    .Call(`_scoper_countSeqsWithInvalidBases_rcpp`, seqs)
}

fastDist_rcpp <- function(seqs) {
    .Call(`_scoper_fastDist_rcpp`, seqs)
}

